//                                               -*- C++ -*-
/**
 *  @file  KFold.cxx
 *  @brief K-Fold cross validation
 *
 *  (C) Copyright 2005-2011 EDF-EADS-Phimeca
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  @author: $LastChangedBy: schueller $
 *  @date:   $LastChangedDate: 2010-11-09 13:44:00 +0100 (Tue, 09 Nov 2010) $
 *  Id:      $Id: KFold.hxx 1649 2010-11-09 12:44:00Z dutka $
 */

#include "Exception.hxx"
#include "PersistentObjectFactory.hxx"
#include "ResourceMap.hxx"
#include "KFold.hxx"
#include "PenalizedLeastSquaresAlgorithm.hxx"

namespace OpenTURNS {

  namespace Base {

    namespace Algo {

      CLASSNAMEINIT(KFold);

      typedef Base::Type::Matrix                     Matrix;
      typedef Base::Type::SymmetricMatrix            SymmetricMatrix;
      typedef Base::Type::Indices                    Indices;
      typedef Base::Type::NumericalPoint             NumericalPoint;
      typedef Base::Common::InvalidArgumentException InvalidArgumentException;
      typedef Base::Func::NumericalMathFunction      NumericalMathFunction;
      typedef Base::Common::ResourceMap              ResourceMap;

      static Base::Common::Factory<KFold> RegisteredFactory("KFold");

      const UnsignedLong KFold::DefaultK = ResourceMap::GetAsUnsignedLong( "KFold-DefaultK" );

      /* Default constructor */
      KFold::KFold( const UnsignedLong k )
        : FittingAlgorithmImplementation()
      {
        setK( k );
      }

      /* Virtual constructor */
      KFold * KFold::clone() const
      {
        return new KFold( *this );
      }

      /* String converter */
      String KFold::__repr__() const
      {
        return OSS() << "class=" << GetClassName();
      }

      /* Perform cross-validation */
      NumericalScalar KFold::run(const NumericalSample & x,
                                 const NumericalSample & y,
                                 const Basis & basis) const
      {
        const UnsignedLong dimension( x.getDimension() );
        const UnsignedLong sampleSize( x.getSize() );
        const NumericalScalar variance( y.computeVariancePerComponent()[0] );

        if ( y.getDimension() != 1 ) throw InvalidArgumentException( HERE ) << "Output sample should be unidimensional (dim=" << y.getDimension() << ").";
        if ( y.getSize() != sampleSize ) throw InvalidArgumentException( HERE ) << "Samples should be equally sized (in=" << sampleSize << " out="<<y.getSize() << ").";
        if ( k_ >= sampleSize ) throw InvalidArgumentException( HERE ) << "K (" << k_ << ") should be < size (" << sampleSize << ").";
        if ( variance <= 0.0 ) throw InvalidArgumentException( HERE ) << "Null output sample variance.";

        NumericalScalar quadraticResidual(0.0);

        // the size of a subsample
        const UnsignedLong testSize( sampleSize / k_ );
        const UnsignedLong trainingSize( ( k_ - 1 ) * testSize );

        // We build the test sample by selecting one over k points of the given samples up to the test size, with a varying initial index
        // i is the initial index
        for ( UnsignedLong i = 0; i < k_; ++ i )
          {

            // build training/test samples
            NumericalSample xTrain( 0, dimension );
            NumericalSample yTrain( 0, 1 );
            NumericalSample xTest( 0, dimension );
            NumericalSample yTest( 0, 1 );

            for ( UnsignedLong j = 0; j < ( k_ * testSize ); ++ j )
              {
                if ( (j % k_) != i )
                  {
                    xTrain.add( x[j] );
                    yTrain.add( y[j] );
                  }
                else
                  {
                    xTest.add( x[j] );
                    yTest.add( y[j] );
                  }
              } // Partitioning loop

            // perform LS regression on the current basis
            PenalizedLeastSquaresAlgorithm penalizedLeastSquaresAlgorithm( xTrain, yTrain, basis );
            penalizedLeastSquaresAlgorithm.run();
            NumericalPoint coefficientsTrain( penalizedLeastSquaresAlgorithm.getCoefficients() );

            // evaluate on the test sample
            NumericalMathFunction metamodelTrain( basis, coefficientsTrain );
            NumericalSample yHatTest(metamodelTrain(xTest));
            // The empirical error is the normalized L2 error
            for ( UnsignedLong j = 0; j < testSize; ++ j ) quadraticResidual += pow( yTest[j][0] - yHatTest[j][0], 2.0 );
          }
          
        const NumericalScalar empiricalError( quadraticResidual / ( testSize * k_ ) );
          
        return empiricalError / variance;
      }


      /* Method save() stores the object through the StorageManager */
      void KFold::save(StorageManager::Advocate & adv) const
      {
        FittingAlgorithmImplementation::save(adv);
        adv.saveAttribute( "k_", k_ );
      }


      /* Method load() reloads the object from the StorageManager */
      void KFold::load(StorageManager::Advocate & adv)
      {
        FittingAlgorithmImplementation::load(adv);
        adv.loadAttribute( "k_", k_ );
      }


      /* P accessor */
      void KFold::setK(const UnsignedLong k)
      {
        if ( k < 1 )
          throw InvalidArgumentException( HERE ) << "KFold k parameter should be > 0";
        k_ = k;
      }


      UnsignedLong KFold::getK() const
      {
        return k_;
      }


    } /* namespace Algo */
  } /* namespace Base */
} /* namespace OpenTURNS */
