//                                               -*- C++ -*-
/**
 *  @file  SymmetricMatrix.cxx
 *  @brief SymmetricMatrix implements the classical mathematical symmetric matrix
 *
 *  (C) Copyright 2005-2011 EDF-EADS-Phimeca
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  @author: $LastChangedBy: schueller $
 *  @date:   $LastChangedDate: 2011-06-30 10:19:34 +0200 (Thu, 30 Jun 2011) $
 *  Id:      $Id: SymmetricMatrix.cxx 1972 2011-06-30 08:19:34Z schueller $
 */
#include "SymmetricMatrix.hxx"

namespace OpenTURNS
{

  namespace Base
  {

    namespace Type
    {

      CLASSNAMEINIT(SymmetricMatrix);

      /* Default constructor */
      SymmetricMatrix::SymmetricMatrix()
        : SquareMatrix(0),
          hasBeenSymmetrized_(false)
      {
        // Nothing to do
      }


      /* Constructor with size (dim, which is the same for nbRows_ and nbColumns_ )*/
      SymmetricMatrix::SymmetricMatrix(const UnsignedLong dim)
        : SquareMatrix(dim),
          hasBeenSymmetrized_(false)
      {
        // Nothing to do
      }

      /* Constructor from external collection */
      /* If the dimensions of the matrix and of the collection */
      /* do not match, either the collection is truncated */
      /* or the rest of the matrix is filled with zeros */
      SymmetricMatrix::SymmetricMatrix(const UnsignedLong dim,
                                       const Collection<NumericalScalar> & elementsValues)
        : SquareMatrix(dim, elementsValues),
          hasBeenSymmetrized_(false)
      {
        // Nothing to do
      }


      /* Constructor with implementation */
      SymmetricMatrix::SymmetricMatrix(const Implementation & i)
        : SquareMatrix(i) ,
          hasBeenSymmetrized_(false)
      {
        // Nothing to do
      }


      /* Check if the internal representation is really symmetric */
      void SymmetricMatrix::checkSymmetry() const
      {
        if (!hasBeenSymmetrized_)
          {
            getImplementation()->symmetrize();
            hasBeenSymmetrized_ = true;
          }
      }

      /* String converter */
      String SymmetricMatrix::__repr__() const
      {
        checkSymmetry();
        return OSS() << "class=" << getClassName()
                     << " dimension=" << this->getDimension()
                     << " implementation=" << getImplementation()->__repr__();
      }

      String SymmetricMatrix::__str__(const String & offset) const
      {
        checkSymmetry();
        return SquareMatrix::__str__();
      }

      /* Operator () gives access to the elements of the matrix (to modify these elements) */
      /* The element of the matrix is designated by its row number i and its column number j */
      /* the first element of the matrix is m(0,0) */
      NumericalScalar & SymmetricMatrix::operator() (const UnsignedLong i,
                                                     const UnsignedLong j) /* throw(InvalidDimensionException) */
      {
        copyOnWrite();
        hasBeenSymmetrized_ = false;

        return (i < j) ? (*getImplementation())(i, j) : (*getImplementation())(j, i) ;
      }

      /* Operator () gives access to the elements of the matrix (read only) */
      /* The element of the matrix is designated by its row number i and its column number j */
      const NumericalScalar & SymmetricMatrix::operator() (const UnsignedLong i,
                                                           const UnsignedLong j)  const /* throw(InvalidDimensionException) */
      {
        return (i < j) ? (*getImplementation())(i, j) : (*getImplementation())(j, i) ;
      }

      /* SymmetricMatrix transpose */
      SymmetricMatrix SymmetricMatrix::transpose () const
      {
        return *this;
      }

      /* SymmetricMatrix additions (must have the same dimensions) */
      Matrix SymmetricMatrix::operator + (const Matrix & m) const /* throw(InvalidDimensionException) */
      {
        checkSymmetry();
        return Matrix(new MatrixImplementation(*getImplementation() + *(m.getImplementation()) ));
      }

      SquareMatrix SymmetricMatrix::operator + (const SquareMatrix & m) const /* throw(InvalidDimensionException) */
      {
        checkSymmetry();
        return SquareMatrix(new MatrixImplementation(*getImplementation() + *(m.getImplementation()) ));
      }

      SymmetricMatrix SymmetricMatrix::operator + (const SymmetricMatrix & m) const /* throw(InvalidDimensionException) */
      {
        return SymmetricMatrix(new MatrixImplementation(*getImplementation() + *(m.getImplementation()) ));
      }

      /* SymmetricMatrix substractions (must have the same dimensions) */
      Matrix SymmetricMatrix::operator - (const Matrix & m) const /* throw(InvalidDimensionException) */
      {
        checkSymmetry();
        return SymmetricMatrix(new MatrixImplementation(*getImplementation() - *(m.getImplementation()) ));
      }

      SquareMatrix SymmetricMatrix::operator - (const SquareMatrix & m) const /* throw(InvalidDimensionException) */
      {
        checkSymmetry();
        return SymmetricMatrix(new MatrixImplementation(*getImplementation() - *(m.getImplementation()) ));
      }

      SymmetricMatrix SymmetricMatrix::operator - (const SymmetricMatrix & m) const /* throw(InvalidDimensionException) */
      {
        return SymmetricMatrix(new MatrixImplementation(*getImplementation() - *(m.getImplementation()) ));
      }

      /* SymmetricMatrix multiplications (must have consistent dimensions) */
      Matrix SymmetricMatrix::operator * (const Matrix & m) const /* throw(InvalidDimensionException) */
      {
        return SymmetricMatrix(new MatrixImplementation(getImplementation()->symProd(*(m.getImplementation()), 'L') ));
      }

      SquareMatrix SymmetricMatrix::operator * (const SquareMatrix & m) const /* throw(InvalidDimensionException) */
      {
        return SymmetricMatrix(new MatrixImplementation(getImplementation()->symProd(*(m.getImplementation()), 'L') ));
      }

      SymmetricMatrix SymmetricMatrix::operator * (const SymmetricMatrix & m) const /* throw(InvalidDimensionException) */
      {
        m.checkSymmetry();
        return SymmetricMatrix(new MatrixImplementation(getImplementation()->symProd(*(m.getImplementation()), 'L') ));
      }

      SymmetricMatrix SymmetricMatrix::operator * (const IdentityMatrix & m) const /* throw(InvalidDimensionException) */
      {
        return *this;
      }


      /* Multiplication with a NumericalPoint (must have consistent dimensions) */
      NumericalPoint SymmetricMatrix::operator * (const NumericalPoint & pt) const /* throw(InvalidDimensionException) */
      {
        return getImplementation()->symVectProd(pt) ;
      }


      /* Multiplication with a NumericalScalar */
      SymmetricMatrix SymmetricMatrix::operator * (const NumericalScalar & s) const
      {
        return SymmetricMatrix(new MatrixImplementation(*getImplementation() * s ));
      }

      /* Division by a NumericalScalar*/
      SymmetricMatrix SymmetricMatrix::operator / (const NumericalScalar & s) const /* throw(InvalidArgumentException) */
      {
        return SymmetricMatrix(new MatrixImplementation(*getImplementation() / s ));
      }

      /* SquareMatrix integer power */
      SymmetricMatrix SymmetricMatrix::power(const UnsignedLong n) const
      {
        return SymmetricMatrix(new MatrixImplementation(getImplementation()->symPower(n)));
      }

      /* Resolution of a linear system */
      NumericalPoint SymmetricMatrix::solveLinearSystem(const NumericalPoint & b,
							const Bool keepIntact)
      {
        return getImplementation()->solveLinearSystemSym(b, keepIntact);
      }

      Matrix SymmetricMatrix::solveLinearSystem(const Matrix & b,
						const Bool keepIntact)
      {
        return getImplementation()->solveLinearSystemSym(*b.getImplementation(), keepIntact);
      }

      /* Compute determinant */
      NumericalScalar SymmetricMatrix::computeDeterminant(const Bool keepIntact)
      {
        return getImplementation()->computeDeterminantSym(keepIntact);
      }

      /* Compute eigenvalues */
      NumericalPoint SymmetricMatrix::computeEigenValues(const Bool keepIntact)
      {
        return getImplementation()->computeEigenValuesSym(keepIntact);
      }


    } /* namespace Type */
  } /* namespace Base */
} /* namespace OpenTURNS */
