//                                               -*- C++ -*-
/**
 *  @file  t_InverseHaltonSequence_std.cxx
 *  @brief Implementation of the InverseHalton' sequence
 *
 *  (C) Copyright 2005-2011 EDF-EADS-Phimeca
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  @author: $LastChangedBy: schueller $
 *  @date:   $LastChangedDate: 2008-05-23 13:46:12 +0200 (ven, 23 mai 2008) $
 *  Id:      $Id: t_InverseHaltonSequence_std.cxx 821 2008-05-23 11:46:12Z schueller $
 */

#include <iostream>
#include <cmath>
#include <ctime>

#include "OT.hxx"
#include "OTtestcode.hxx"
#include "OStream.hxx"
#include "InverseHaltonSequence.hxx"
#include "IndependentCopula.hxx"
#include "ComposedDistribution.hxx"
#include "Uniform.hxx"
#include "CorrelationAnalysis.hxx"

using namespace OT;
using namespace OT::Test;
using namespace OT::Base::Common;
using namespace OT::Base::Type;
using namespace OT::Base::Stat;
using namespace OT::Uncertainty::Algorithm;
using namespace OT::Uncertainty::Model;
using namespace OT::Uncertainty::Distribution;

typedef Collection<Distribution>        DistributionCollection;

int main(int argc, char *argv[])
{
  TESTPREAMBLE;
  OStream fullprint(std::cout);

  try {
    // Create a InverseHalton sequence
    InverseHaltonSequence sequence(15);
    fullprint << sequence << std::endl;

    // Create a numerical sample of the sequence
    NumericalSample inverseHaltonSample(sequence.generate(10));
    fullprint << inverseHaltonSample << std::endl;

    // Create another InverseHalton' sequence of dimension 2 to estimate Pi in [0; 1)^2
    UnsignedLong dimension(2);
    sequence = InverseHaltonSequence(dimension);
    UnsignedLong pointInsideCircle(0);
    UnsignedLong sampleSize(1000);
    for(UnsignedLong i = 0; i < sampleSize; ++i)
      {
        NumericalPoint inverseHaltonPoint(sequence.generate());
        fullprint << inverseHaltonPoint << std::endl;
        if(inverseHaltonPoint.norm() < 1.0)
          ++ pointInsideCircle;
      }
    NumericalScalar probabilityEstimate(1.0 * pointInsideCircle / sampleSize);
    NumericalScalar probability(M_PI / 4.0);
    NumericalScalar relativeError(fabs(probability - probabilityEstimate) / probability);
    fullprint <<"sample size=" << sampleSize << std::endl;
    fullprint <<"relative error to Pi=" << relativeError << std::endl;


    // we create a collection of uniform distributions over [-Pi; Pi[
    DistributionCollection aCollection;
    for(UnsignedLong i = 0; i < dimension; ++i)
      {
        aCollection.add(Uniform(0.0, 1.0));
      }

    // we create an independent copula
    IndependentCopula aCopula(aCollection.getSize());
    aCopula.setName("an independent copula");

    // we create one distribution object
    ComposedDistribution aDistribution(aCollection, Copula(aCopula));
    aDistribution.setName("a uniform distribution");
  }




  catch (TestFailed & ex) {
    std::cerr << ex << std::endl;
    return ExitCode::Error;
  }

  return ExitCode::Success;
}
